@@ -49,7 +49,7 @@
 
     This program is to demonstrate how to use the m17n GUI API.
     Although m17n-edit directly uses the GUI API, the API is mainly
-    for toolkit libraries or to implement XOM (X Outout Method), not
+    for toolkit libraries or to implement XOM (X Output Method), not
     for direct use from application programs.
 */
 /***ja
@@ -81,7 +81,7 @@
 
     Υץ m17n GUI API λȤ򼨤ΤǤ롣m17n-edit
     ľ GUI API ȤäƤ뤬 API ϼ˥ġ륭åȥ饤֥
-    XOM (X Outout Method) μѤǤꡢץꥱץ
+    XOM (X Output Method) μѤǤꡢץꥱץ
     फľܤѤտޤƤʤ
 */
 
@@ -120,7 +120,7 @@
 #include <X11/Xaw/SmeLine.h>
 #include <X11/Xaw/MenuButton.h>
 
-#define VERSION "1.2.0"
+#define VERSION "1.3.4"
 
 /* Global variables.  */
 
@@ -865,8 +865,7 @@ show_cursor (XtPointer client_data)
       else
 	{
 	  XtSetArg (arg[0], XtNborderWidth, 1);
-	  XtSetArg (arg[1], XtNlabel,
-		    msymbol_name (msymbol_get (sym, Mlanguage)));
+	  XtSetArg (arg[1], XtNlabel, mlanguage_name (sym));
 	  XtSetValues (CurLangWidget, arg, 2);
 	}
       XtSetValues (CurLangWidget, arg, 2);
@@ -1282,7 +1281,7 @@ ButtonProc (Widget w, XEvent *event, String *str, Cardinal *num)
     {
       MText *produced = mtext ();
 
-          minput_lookup (current_input_context, Mnil, NULL, produced);
+      minput_lookup (current_input_context, Mnil, NULL, produced);
       if (mtext_len (produced) > 0)
 	{
 	  insert_chars (produced);
@@ -2185,7 +2184,7 @@ surrounding_text_handler (MInputContext *ic, MSymbol command)
     {
       int len = (int) mplist_value (ic->plist);
       int pos;
-      MText *surround = NULL;
+      MText *surround;
 
       if (len < 0)
 	{
@@ -2201,11 +2200,10 @@ surrounding_text_handler (MInputContext *ic, MSymbol command)
 	    pos = nchars;
 	  surround = mtext_duplicate (mt, cursor.from, pos);
 	}
-      if (surround)
-	{
-	  mplist_set (ic->plist, Mtext, surround);
-	  m17n_object_unref (surround);
-	}
+      else
+	surround = mtext ();
+      mplist_set (ic->plist, Mtext, surround);
+      m17n_object_unref (surround);
     }
   else if (command == Minput_delete_surrounding_text)
     {
@@ -2247,8 +2245,8 @@ compare_input_method (const void *elt1, const void *elt2)
     return 1;
   if (im2->language == Mt)
     return -1;
-  lang1 = msymbol_get (im1->language, Mlanguage);
-  lang2 = msymbol_get (im2->language, Mlanguage);
+  lang1 = mlanguage_name (im1->language);
+  lang2 = mlanguage_name (im2->language);
   return strcmp (msymbol_name (lang1), msymbol_name (lang2));
 }
 
@@ -2506,7 +2504,8 @@ create_menu_button (Widget top, Widget parent, Widget left, char *button_name,
   XtSetArg (arg[3], XtNhighlightThickness, 1);
   XtSetArg (arg[4], XtNleft, XawChainLeft);
   XtSetArg (arg[5], XtNright, XawChainLeft);
-  i = 6;
+  XtSetArg (arg[6], XtNinternational, True);
+  i = 7;
   if (left)
     XtSetArg (arg[i], XtNfromHoriz, left), i++;
   button = XtCreateManagedWidget (button_name, menuButtonWidgetClass, parent,
@@ -2963,7 +2962,7 @@ main (int argc, char **argv)
 
 	if (im->language != Mnil && im->language != Mt)
 	  {
-	    MSymbol sym = msymbol_get (im->language, Mlanguage);
+	    MSymbol sym = mlanguage_name (im->language);
 	    if (sym == Mnil)
 	      name1 = msymbol_name (im->language);
 	    else
@@ -3090,7 +3089,7 @@ main (int argc, char **argv)
 	    MSymbol fullname;
 
 	    if (sym != Mnil
-		&& ((fullname = msymbol_get (sym, Mlanguage)) != Mnil))
+		&& ((fullname = mlanguage_name (sym)) != Mnil))
 	      {
 		char *name = msymbol_name (fullname);
 		char c = name[0];
