@@ -34,7 +34,7 @@
     The m17n library utilizes this tendency to store characters and
     their attribute values efficiently in an object called @e
     Chartable.  Although a chartable object is not a simple array,
-    application programs can handle chartables as if they were arrays.
+    application programs can handle a chartable as if it is an array.
     Attribute values of a character can be obtained by accessing a
     Chartable for the attribute with the character code of the
     specified character.
@@ -46,17 +46,14 @@
 
     @brief ʸơ֥Ȥ˴ؤ API.
 
-    m17n 饤֥꤬ʸζ֤ϹǤ뤿ᡢʸξñ
-    ˳Ǽ褦ȤȡϵˤʤꤹŪ
-    롣̾ɬפȤʤʸˤĤƤξϡϰϤʸ
-    ˤΤդƤ뤳Ȥ¿ʸ˴ؤƾ󤬤ˤ⡢Ϣ
-    ³ʸɤʸƱĤȤ¿
+    m17n 饤֥꤬ʸζ֤ϹǤ뤿ᡢʸξñ˳Ǽ褦ȤȡϵˤʤꤹŪǤ롣
+    ̾ɬפȤʤʸˤĤƤξϡϰϤʸˤΤդƤ뤳Ȥ¿
+    ʸ˴ؤƾ󤬤ˤ⡢Ϣ³ʸɤʸƱĤȤ¿
 
     Τ褦ʷѤʸȤղþΨŪ˳Ǽ뤿ˡ
-    m17n 饤֥ @e ʸơ֥ (chartable) ȸƤ֥֥Ȥ
-    Ѥ롣ʸơ֥ǤϤʤץꥱץ
-    ʸơ֥ΰȤưȤǤ롣ʸˤĤƤ
-    ξϡξʸơ֥򤽤ʸΥɤǰ
+    m17n 饤֥ @e ʸơ֥ (chartable) ȸƤ֥֥ȤѤ롣
+    ʸơ֥ǤϤʤץꥱץʸơ֥ΰȤưȤǤ롣
+    ʸˤĤƤξϡξʸơ֥򤽤ʸΥɤǰ
     롣  
 
     ʸơ֥ϴ֥ȤǤ롣 */
@@ -603,14 +600,13 @@ dump_sub_chartab (MSubCharTable *table, void *default_value,
 int
 mchartable__init ()
 {
-  chartable_table.count = 0;
+  M17N_OBJECT_ADD_ARRAY (chartable_table, "Chartable");
   return 0;
 }
 
 void
 mchartable__fini ()
 {
-  mdebug__report_object ("Chartable", &chartable_table);
 }
 
 void *
@@ -662,13 +658,12 @@ MSymbol Mchar_table;
 
     ؿ mchartable () ϥ $KEY ǤΥǥեͤ 
     $DEFAULT_VALUE Ǥ뿷ʸơ֥롣⤷ $KEY 
-    ǤСΥơ֥Ǥϡʥǥեͤޤơ˴֥
-    Ȥ NULL Τ줫Ǥ롣
+    ǤСΥơ֥Ǥϡʥǥեͤޤơ˴֥Ȥ 
+    NULL Τ줫Ǥ롣
 
     @return
-     mchartable () Ϻ줿ʸơ֥ؤΥݥ
-    ֤Ԥ @c NULL ֤ѿ #merror_code ˥
-    顼ɤꤹ롣  */
+     mchartable () Ϻ줿ʸơ֥ؤΥݥ󥿤֤
+    Ԥ @c NULL ֤ѿ #merror_code ˥顼ɤꤹ롣  */
 
 MCharTable *
 mchartable (MSymbol key, void *default_value)
@@ -702,9 +697,9 @@ mchartable (MSymbol key, void *default_value)
 /***ja
     @brief ʸơ֥ʸ˳Ƥ줿֤ͤ.
 
-    ؿ mchartable_lookup () ʸơ֥ $TABLE ʸ $C ˳
-    Ƥ줿֤ͤ$C ФŪͤʤС$TABLE Υǥե
-    ֤ͤ$C ʸǤʤСmchartable_lookup ()  
+    ؿ mchartable_lookup () ʸơ֥ $TABLE ʸ $C 
+    ˳Ƥ줿֤ͤ$C ФŪͤʤС$TABLE 
+    Υǥե֤ͤ$C ʸǤʤСmchartable_lookup ()  
     @c NULL ֤ѿ #merror_code ˥顼ɤꤹ롣  */
 
 /***
@@ -819,8 +814,8 @@ mchartable_set (MCharTable *table, int c, void *val)
     @return
      mchartable_set_range ()  0 ֤Ǥʤ 
     -1 ֤ѿ #merror_code ˥顼ɤꤹ롣$FROM  
-    $TO 礭Ȥˤϡ mchartable_set_range () ϲ⤻顼
-    ⵯʤ  */
+    $TO 礭Ȥˤϡ mchartable_set_range () 
+    ϲ⤻顼ⵯʤ  */
 
 /***
     @errors
@@ -866,9 +861,9 @@ mchartable_set_range (MCharTable *table, int from, int to, void *val)
 /***ja
     @brief ͤǥեȤȰۤʤʸõ.
 
-    ؿ mchartable_range () ʸơ֥ $TABLE ǡ$TABLE Υǥե
-    ͰʳͤĺǽȺǸʸõ줾 $FROM  
-    $TO ꤹ롣٤ƤʸǥեͤͤȤƻäƤС
+    ؿ mchartable_range () ʸơ֥ $TABLE ǡ$TABLE 
+    ΥǥեͰʳͤĺǽȺǸʸõ줾 $FROM 
+     $TO ꤹ롣٤ƤʸͤȤƥǥեͤȤäƤˤ
     $FROM  $TO  -1ꤹ롣  */
 
 void
@@ -910,19 +905,19 @@ mchartable_range (MCharTable *table, int *from, int *to)
 /***ja
     @brief ʸơ֥ʸФƻδؿƤ.
 
-    ؿ mchartable_map () ϡʸơ֥ $TABLE ʸФƴ
-     $FUNC Ƥ֡$TABLE Ǥͤ $IGNORE ǤʸˤĤ
-    ƤϴؿƤӽФԤʤʤ$IGNORE ʸͤӤ @c == 
-    ԤʤΤǡʸƥݥ󥿤Ȥݤˤդפ롣
+    ؿ mchartable_map () ϡʸơ֥ $TABLE ʸФƴؿ
+    $FUNC Ƥ֡$TABLE Ǥͤ $IGNORE 
+    ǤʸˤĤƤϴؿƤӽФԤʤʤ$IGNORE ʸͤӤ 
+    @c == ǹԤʤΤǡʸƥݥ󥿤Ȥݤˤդפ롣
 
-    mchartable_map () ϡʸȤ $FUNC Ƥ֤ΤǤϤʤؿ
-    ӽФβŬ褦Ȥ롣ʤϢ³ʸƱͤ
-    äƤˤϡʸΤޤȤޤΤˤĤư٤δؿƤӽ
+    mchartable_map () ϡʸȤ $FUNC 
+    Ƥ֤ΤǤϤʤؿƤӽФβŬ褦Ȥ롣
+    ʤϢ³ʸƱͤäƤˤϡʸΤޤȤޤΤˤĤư٤δؿƤӽ
     Ԥʤʤ
 
     ʸΤޤȤޤ礭ˤ餺$FUNC  $FROM, $TO, $VAL, $ARG 
-    ΣǸƤФ롣$FROM  $TO ξüޤˤ $VAL ͤȤ
-    ʸϰϤ򼨤$ARG  $FUNC_ARG ΤΤǤ롣
+    ΣǸƤФ롣$FROM  $TO ξüޤˤ $VAL 
+    ͤȤƻʸϰϤ򼨤$ARG  $FUNC_ARG ΤΤǤ롣
 
     @return
     δؿϾ0֤  */
@@ -957,8 +952,8 @@ mchartable_map (MCharTable *table, void *ignore,
 /***ja
     @brief ʸơ֥פ.
 
-    ؿ mdebug_dump_chartab () ʸơ֥ $TABLE  stderr ˿ʹ
-    ˲ɤʷǰ롣$INDENT ϣܰʹߤΥǥȤꤹ롣
+    ؿ mdebug_dump_chartab () ʸơ֥ $TABLE  stderr 
+    ˿ʹ֤˲ɤʷǰ롣$INDENT ϣܰʹߤΥǥȤꤹ롣
 
     @return
     δؿ $TABLE ֤  */
