@@ -1,5 +1,5 @@
 /* font.c -- font module.
-   Copyright (C) 2003, 2004
+   Copyright (C) 2003, 2004, 2005, 2006
      National Institute of Advanced Industrial Science and Technology (AIST)
      Registration Number H15PRO112
 
@@ -336,6 +336,7 @@
 #include "symbol.h"
 #include "plist.h"
 #include "charset.h"
+#include "language.h"
 #include "internal-gui.h"
 #include "font.h"
 #include "face.h"
@@ -663,55 +664,18 @@ OTF_tag (char *name)
 			  | (! p[2] ? 0
 			     : (p[2] << 8) | p[3]))));
 }
+
+void
+OTF_tag_name (OTF_Tag tag, char *name)
+{
+  name[0] = (char) (tag >> 24);
+  name[1] = (char) ((tag >> 16) & 0xFF);
+  name[2] = (char) ((tag >> 8) & 0xFF);
+  name[3] = (char) (tag & 0xFF);
+  name[4] = '\0';
+}
 #endif	/* not HAVE_OTF */
 
-static MPlist *otf_script_list;
-
-static int
-load_otf_script_list ()
-{
-  MDatabase *mdb;
-  MPlist *plist, *pl;
-
-  otf_script_list = mplist ();
-  mdb = mdatabase_find (msymbol ("standard"), Mscript, msymbol ("otf"), Mnil);
-  if (! mdb
-      || ! (plist = mdatabase_load (mdb)))
-    MERROR (MERROR_FONT, -1);
-  MPLIST_DO (pl, plist)
-    {
-      MPlist *p;
-      MSymbol script, otf_script;
-      OTF_Tag tag;
-
-      if (! MPLIST_PLIST_P (pl))
-	continue;
-      p = MPLIST_PLIST (pl);
-      if (! MPLIST_SYMBOL_P (p))
-	continue;
-      script = MPLIST_SYMBOL (p);
-      p = MPLIST_NEXT (p);
-      if (! MPLIST_SYMBOL_P (p))
-	continue;
-      otf_script = MPLIST_SYMBOL (p);
-      tag = OTF_tag (MSYMBOL_NAME (otf_script));
-      mplist_push (otf_script_list, script, (void *) tag);
-    }
-  M17N_OBJECT_UNREF (plist);
-  return 0;
-}
-
-static MSymbol
-find_script_from_otf_tag (OTF_Tag tag)
-{
-  MPlist *plist;
-
-  if (! otf_script_list)
-    load_otf_script_list ();
-  plist = mplist_find_by_value (otf_script_list, (void *) tag);
-  return (plist ? MPLIST_KEY (plist) : Mnil);
-}
-
 /* XLFD parser/generator */
 
 /** Indices to each field of split font name.  */
@@ -744,6 +708,7 @@ xlfd_parse_name (const char *name, MFont *font)
   char copy[513];
   int i;
   char *p;
+  MSymbol sym;
 
   if (name[0] != '-')
     return -1;
@@ -793,9 +758,19 @@ xlfd_parse_name (const char *name, MFont *font)
     size = atoi (field[XLFD_PIXEL]) * 10;
 
   if (field[XLFD_FOUNDRY])
-    mfont__set_property (font, MFONT_FOUNDRY, msymbol (field[XLFD_FOUNDRY]));
+    {
+      sym = msymbol (field[XLFD_FOUNDRY]);
+      if (! sym)
+	sym = msymbol ("Nil");
+      mfont__set_property (font, MFONT_FOUNDRY, sym);
+    }
   if (field[XLFD_FAMILY])
-    mfont__set_property (font, MFONT_FAMILY, msymbol (field[XLFD_FAMILY]));
+    {
+      sym = msymbol (field[XLFD_FAMILY]);
+      if (! sym)
+	sym = msymbol ("Nil");
+      mfont__set_property (font, MFONT_FAMILY, sym);
+    }
   if (field[XLFD_WEIGHT])
     mfont__set_property (font, MFONT_WEIGHT, msymbol (field[XLFD_WEIGHT]));
   if (field[XLFD_SLANT])
@@ -828,6 +803,7 @@ xlfd_unparse_name (MFont *font, int full_xlfd)
   int len, i;
   char spacing;
   int size, resy;
+  int all_nil = 1;
 
   prop[0] = (MSymbol) mfont_get_prop (font, Mfoundry);
   prop[1] = (MSymbol) mfont_get_prop (font, Mfamily);
@@ -842,6 +818,7 @@ xlfd_unparse_name (MFont *font, int full_xlfd)
 	{
 	  str[i] = msymbol_name (prop[i]);
 	  len += strlen (str[i]);
+	  all_nil = 0;
 	}
       else
 	{
@@ -867,7 +844,14 @@ xlfd_unparse_name (MFont *font, int full_xlfd)
   size = font->size;
   if (size >= 0)
     {
-      if ((size % 10) < 5)
+      if (font->multiple_sizes)
+	{
+	  for (size = 0; size < 24; size++)
+	    if (font->size & (1 << size))
+	      break;
+	  size += 6;
+	}
+      else if ((size % 10) < 5)
 	size /= 10;
       else
 	size = size / 10 + 1;
@@ -886,6 +870,8 @@ xlfd_unparse_name (MFont *font, int full_xlfd)
 		 str[0], str[1], str[2], str[3], str[4], str[5],
 		 size, resy, resy, spacing, str[6]);
     }
+  else if (all_nil && size == 0)
+    sprintf (name, "*");
   else
     {
       char *p = name;
@@ -925,7 +911,21 @@ font_score (MFont *font, MFont *request)
 	{
 	  if (font->size && request->size)
 	    {
-	      val = font->size - request->size;
+	      if (font->multiple_sizes)
+		{
+		  int j, closest = 23;
+
+		  for (j = 23; j >= 0; j--)
+		    if (font->size & (1 << j))
+		      {
+			closest = j;
+			if (request->size >= (j + 6) * 10)
+			  break;
+		      }
+		  val = request->size - (closest + 6) * 10;
+		}
+	      else
+		val = font->size - request->size;
 	      if (val)
 		{
 		  if (val < 0)
@@ -968,6 +968,82 @@ font_score (MFont *font, MFont *request)
   return score;
 }
 
+static MSymbol
+merge_capability (MSymbol capability, MSymbol key, MSymbol val, int overwrite)
+{
+  MFontCapability *cap = NULL;
+  char *lang = NULL, *script = NULL, *otf = NULL, *buf, *p;
+  int lang_len = 0, script_len = 0, otf_len = 0;
+
+  if (key == Mlanguage)
+    lang = MSYMBOL_NAME (val), lang_len = MSYMBOL_NAMELEN (val) + 6;
+  else if (key == Mscript)
+    script = MSYMBOL_NAME (val), script_len = MSYMBOL_NAMELEN (val) + 7;
+  else if (key == Motf)
+    otf = MSYMBOL_NAME (val), otf_len = MSYMBOL_NAMELEN (val) + 5;
+  else
+    return capability;
+
+  if (capability != Mnil)
+    {
+      cap = mfont__get_capability (capability);
+      if (! overwrite)
+	{
+	  if (cap->language)
+	    lang = NULL;
+	  if (cap->script)
+	    script = NULL;
+	  if (cap->script_tag)
+	    otf = NULL;
+	  if (! lang && !script && !otf)
+	    return capability;
+	}
+    }
+
+  if (! lang && cap && cap->language)
+    {
+      lang_len = MSYMBOL_NAMELEN (cap->language);
+      lang = MSYMBOL_NAME (cap->language);
+    }
+  if (! script && cap && cap->script != Mnil)
+    {
+      script_len = MSYMBOL_NAMELEN (cap->script);
+      script = MSYMBOL_NAME (cap->script);
+    }
+  if (! otf && cap && cap->script_tag)
+    {
+      int i;
+
+      otf_len = 4;			/* for script_tag */
+      if (cap->langsys_tag)
+	otf_len += 5;		/* for "/XXXX */
+      for (i = 0; i < MFONT_OTT_MAX; i++)
+	if (cap->features[i].str)
+	  otf_len += strlen (cap->features[i].str) + 1; /* for "[=+]..." */
+      otf = p = alloca (otf_len + 1);
+      OTF_tag_name (cap->script_tag, otf);
+      p += 4;
+      if (cap->langsys_tag)
+	{
+	  *p++ = '/';
+	  OTF_tag_name (cap->langsys_tag, p);
+	  p += 4;
+	}
+      if (cap->features[MFONT_OTT_GSUB].str)
+	p += sprintf (p, "=%s", cap->features[MFONT_OTT_GSUB].str);
+      if (cap->features[MFONT_OTT_GPOS].str)
+	p += sprintf (p, "=%s", cap->features[MFONT_OTT_GSUB].str);
+    }
+  buf = p = alloca (lang_len + script_len + otf_len + 1);
+  if (lang_len)
+    p += sprintf (p, ":lang=%s", lang);
+  if (script_len)
+    p += sprintf (p, ":script=%s", script);
+  if (otf_len)
+    p += sprintf (p, ":otf=%s", otf);
+  return msymbol (buf);
+}
+
 
 /* Internal API */
 
@@ -1018,6 +1094,8 @@ mfont__init ()
   Municode_full = msymbol ("unicode-full");
   Mapple_roman = msymbol ("apple-roman");
 
+  Motf = msymbol ("otf");
+
   /* The first entry of each mfont__property_table must be Mnil so
      that actual properties get positive numeric numbers.  */
   for (i = 0; i <= MFONT_REGISTRY; i++)
@@ -1134,11 +1212,6 @@ mfont__fini ()
       M17N_OBJECT_UNREF (font_encoding_list);
       font_encoding_list = NULL;
     }
-  if (otf_script_list)
-    {
-      M17N_OBJECT_UNREF (otf_script_list);
-      otf_script_list = NULL;
-    }
 
   for (i = 0; i <= MFONT_REGISTRY; i++)
     MLIST_FREE1 (&mfont__property_table[i], names);
@@ -1195,8 +1268,17 @@ int
 mfont__match_p (MFont *font, MFont *spec, int prop)
 {
   if (spec->capability != font->capability
-      && spec->capability != Mnil && font->capability != Mnil)
-    return 0;
+      && spec->capability != Mnil)
+    {
+      MRealizedFont *rfont;
+
+      if (font->type != MFONT_TYPE_REALIZED)
+	return (font->capability == Mnil);
+      rfont = (MRealizedFont *) font;
+      return (rfont->driver->check_capability
+	      && (rfont->driver->check_capability (rfont, spec->capability)
+		  >= 0));
+    }
   if (spec->file != font->file
       && spec->file != Mnil && font->file != Mnil)
     return 0;
@@ -1260,6 +1342,7 @@ mfont__set_spec_from_face (MFont *spec, MFace *face)
     mfont__set_property (spec, i, face->property[i]);
   spec->property[MFONT_REGISTRY] = 0;
   spec->property[MFONT_RESY] = 0;
+  spec->multiple_sizes = 0;
   spec->size = (int) (face->property[MFACE_SIZE]);
   spec->type = MFONT_TYPE_SPEC;
   spec->source = MFONT_SOURCE_UNDECIDED;
@@ -1624,9 +1707,7 @@ free_font_capability (void *object)
 {
   MFontCapability *cap = object;
   
-  if (cap->lang)
-    free (cap->lang);
-  if (cap->script)
+  if (cap->script_tag)
     {
       int i;
       for (i = 0; i < MFONT_OTT_MAX; i++)
@@ -1659,22 +1740,29 @@ mfont__get_capability (MSymbol sym)
     {
       if (*str++ != ':')
 	continue;
-      if (str[0] == 'o' && str[1] == 't' && str[2] == 'f' && str[3] == '=')
+      if (str[0] == 'o' && strncmp (str + 1, "tf=", 3) == 0)
 	{
+	  char *beg;
+	  MSymbol sym;
 	  int i;
 
 	  str += 4;
+	  beg = str;
 	  for (i = 0, p = str; i < 4 && p < endp; i++, p++);
 	  if (i < 4)
 	    break;
+	  sym = msymbol__with_len (str, 4);
+	  cap->script = mscript__from_otf_tag (sym);
+	  if (cap->script == Mnil)
+	    break;
 	  cap->script_tag = OTF_tag (str);
-	  cap->script = find_script_from_otf_tag (cap->script_tag);
 	  if (*p == '/')
 	    {
 	      for (i = 0, str = ++p; i < 4 && p < endp; i++, p++);
 	      if (i < 4)
 		{
 		  cap->script = Mnil;
+		  cap->script_tag = 0;
 		  break;
 		}
 	      cap->langsys_tag = OTF_tag (str);
@@ -1739,38 +1827,18 @@ mfont__get_capability (MSymbol sym)
 		cap->features[i].tags = malloc (sizeof (OTF_Tag));
 		cap->features[i].tags[0] = 0;
 	      }
+	  cap->otf = msymbol__with_len (beg, p - beg);
 	  str = p;
 	}
-      else if (str[0] == 'l' && str[1] == 'a' && str[2] == 'n' && str[3] == 'g'
-	  && str[4] == '=')
+      else if (str[0] == 'l' && strncmp (str + 1, "ang=", 4) == 0)
 	{
-	  int count;
-
 	  str += 5;
-	  for (p = str, count = 2; p < endp && *p != ':'; p++)
-	    if (*p == ',')
-	      count++;
-	  MTABLE_MALLOC (cap->lang, count, MERROR_FONT);
-	  for (p = str, count = 0; p < endp && *p != ':'; p++)
-	    if (*p == ',')
-	      {
-		MSymbol lang = msymbol__with_len (str, p - str), sym;
-
-		if (msymbol_get (lang, Miso639_2))
-		  cap->lang[count++] = lang;
-		else if ((sym = msymbol_get (lang, Miso639_1)) != Mnil)
-		  cap->lang[count++] = sym;
-		else if (msymbol_get (lang, Mlanguage))
-		  cap->lang[count++] = lang;
-		str = p + 1;
-	      }
+	  for (p = str; p < endp && *p != ':'; p++);
 	  if (str < p)
-	    cap->lang[count++] = msymbol__with_len (str, p - str);
-	  cap->lang[count] = Mnil;
+	    cap->language = msymbol__with_len (str, p - str);
 	  str = p;
 	}
-      else if (str[0] == 's' && str[1] == 'c' && str[2] == 'r' && str[3] == 'i'
-	       && str[4] == 'p' && str[5] == 't' && str[6] == '=')
+      else if (str[0] == 's' && strncmp (str + 1, "cript=", 6) == 0)
 	{
 	  str += 7;
 	  for (p = str; p < endp && *p != ':'; p++);
@@ -1950,6 +2018,14 @@ MSymbol Mregistry;
 
 MSymbol Msize;
 
+/***en
+    @brief Key of font property specifying file name.
+
+    The variable #Mfontfile is a symbol of name <tt>"fontfile"</tt>
+    and is used as a key of font property.  The property value must be
+    a symbol whose name is a font file name.  */ 
+MSymbol Motf;
+
 /***en
     @brief Key of font property specifying file name.
 
@@ -2239,23 +2315,23 @@ mfont_copy (MFont *font)
     The mfont_get_prop () function gets the value of $KEY property of
     font $FONT.  $KEY must be one of the following symbols:
 
-	@c Mfamily, @c Mweight, @c Mstyle, @c Mstretch, @c Madstyle,
-	@c Mregistry, @c Msize, @c Mresolution, @c Mspacing.
+	@c Mfoundry, @c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
+	@c Madstyle, @c Mregistry, @c Msize, @c Mresolution, @c Mspacing.
 
     If $FONT is a return value of mfont_find (), $KEY can also be one
     of the following symbols:
 
 	#Mfont_ascent, #Mfont_descent, #Mmax_advance.
 
-    @return If $KEY is @c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
-    @c Madstyle, @c Mregistry, or @c Mspacing, this function returns
-    the corresponding value as a symbol.  If the font does not have
-    $KEY property, it returns @c Mnil.  If $KEY is @c Msize, @c
-    Mresolution, #Mfont_ascent, Mfont_descent, or #Mmax_advance, this
-    function returns the corresponding value as an integer.  If the
-    font does not have $KEY property, it returns 0.  If $KEY is
-    something else, it returns @c NULL and assigns an error code to
-    the external variable #merror_code.  */
+    @return If $KEY is @c Mfoundry, @c Mfamily, @c Mweight, @c Mstyle,
+    @c Mstretch, @c Madstyle, @c Mregistry, or @c Mspacing, this
+    function returns the corresponding value as a symbol.  If the font
+    does not have $KEY property, it returns @c Mnil.  If $KEY is @c
+    Msize, @c Mresolution, #Mfont_ascent, Mfont_descent, or
+    #Mmax_advance, this function returns the corresponding value as an
+    integer.  If the font does not have $KEY property, it returns 0.
+    If $KEY is something else, it returns @c NULL and assigns an error
+    code to the external variable #merror_code.  */
  
 /***ja
     @brief եȤΥץѥƥͤ.
@@ -2264,16 +2340,16 @@ mfont_copy (MFont *font)
     $KEY ǤΤ֤ͤ$KEY ϰʲΥܥΤ줫Ǥʤ
     Фʤʤ
 
-	@c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
+	@c Mfoundry, @c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
 	@c Madstyle, @c Mregistry, @c Msize, @c Mresolution, @c Mspacing.
 
-    @return $KEY  @c Mfamily, @c Mweight, @c Mstyle, @c Mstretch, @c
-    Madstyle, @c Mregistry, @c Mspacing Τ줫ǤС
-    򥷥ܥȤ֤եȤΥץѥƥʤˤ
-    @c Mnil ֤$KEY  @c Msize 뤤 @c Mresolution ξˤϡ
-    ͤͤȤ֤եȤΥץѥƥʤ
-    ˤ 0 ֤$KEY ʳΤΤǤС@c NULL ֤
-    ѿ #merror_code ˥顼ɤꤹ롣  */
+    @return $KEY  @c Mfoundry, @c Mfamily, @c Mweight, @c Mstyle, @c
+    Mstretch, @c Madstyle, @c Mregistry, @c Mspacing Τ줫ǤС
+    ͤ򥷥ܥȤ֤եȤΥץѥƥʤ
+    ˤ@c Mnil ֤$KEY  @c Msize 뤤 @c Mresolution 
+    ˤϡͤͤȤ֤եȤΥץѥƥ
+    ʤˤ 0 ֤$KEY ʳΤΤǤС@c
+    NULL ֤ѿ #merror_code ˥顼ɤꤹ롣  */
 
 void *
 mfont_get_prop (MFont *font, MSymbol key)
@@ -2307,6 +2383,20 @@ mfont_get_prop (MFont *font, MSymbol key)
       int resy = font->property[MFONT_RESY];
       return (void *) resy;
     }
+  if (key == Mlanguage || key == Mscript || key == Motf)
+    {
+      MFontCapability *cap;
+
+      if (! font->capability)
+	return NULL;
+      cap = mfont__get_capability (font->capability);
+      if (key == Mlanguage)
+	return cap->language;
+      if (key == Mscript)
+	return cap->script;
+      return cap->otf;
+    }
+
   if (key == Mfontfile)
     return (void *) font->file;
   if (key == Mspacing)
@@ -2334,22 +2424,26 @@ mfont_get_prop (MFont *font, MSymbol key)
     $KEY and value is $VAL to font $FONT.  $KEY must be one of the
     following symbols:
 
-	@c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
+	@c Mfoundry, @c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
 	@c Madstyle, @c Mregistry, @c Msize, @c Mresolution.
 
     If $KEY is @c Msize or @c Mresolution, $VAL must be an integer.
-    Otherwise, $VAL must be a symbol.  */
+    Otherwise, $VAL must be a symbol of a property value name.  But,
+    if the name is "nil", a symbol of name "Nil" must be
+    specified.  */
  /***ja
     @brief եȤΥץѥƥͤꤹ.
 
-    ؿ mfont_put_prop () ϡե $FONT Υ$KEY 
-    Ǥץѥƥͤ $VAL ꤹ롣$KEY ϰʲΥܥΤ줫Ǥ롣
+    ؿ mfont_put_prop () ϡե $FONT Υ$KEY Ǥץ
+    ƥͤ $VAL ꤹ롣$KEY ϰʲΥܥΤ줫Ǥ롣
 
-	@c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
+	@c Mfoundry, @c Mfamily, @c Mweight, @c Mstyle, @c Mstretch,
 	@c Madstyle, @c Mregistry, @c Msize, @c Mresolution.
 
-    $KEY  @c Msize  @c Mresolution Ǥ $VAL 
-    ͤǤʤƤϤʤʳξ硢$VAL ϥܥǤʤƤϤʤʤ*/
+    $KEY  @c Msize  @c Mresolution Ǥ $VAL ͤǤʤƤ
+    ʤʳξ硢$VAL ϥץѥƥͤ̾ΥܥǤʤ
+    Ϥʤʤ⤷̾ "nil" ξϡ̾ "Nil" Υ
+    ܥǤʤƤϤʤʤ*/
 
 int
 mfont_put_prop (MFont *font, MSymbol key, void *val)
@@ -2378,18 +2472,10 @@ mfont_put_prop (MFont *font, MSymbol key, void *val)
       unsigned resy = (unsigned) val;
       font->property[MFONT_RESY] = resy;
     }
-  else if (key == Mlanguage)
+  else if (key == Mlanguage || key == Mscript || key == Motf)
     {
-      char *langname = MSYMBOL_NAME ((MSymbol) val);
-      int len = MSYMBOL_NAMELEN ((MSymbol) val);
-      
-      if (len <= 3)
-	{
-	  char buf[10];
-
-	  sprintf (buf, ":lang=%s", langname);
-	  font->capability = msymbol (buf);
-	}
+      font->capability = merge_capability (font->capability,
+					   key, (MSymbol) val, 1);
     }
   else if (key == Mfontfile)
     {
@@ -2770,6 +2856,10 @@ mfont_resize_ratio (MFont *font)
     ones that support $LANGUAGE.  $MAXNUM, if greater than 0, limits
     the number of fonts.
 
+    $LANGUAGE argument exists just for backward compatibility, and the
+    use is deprecated.  Use #Mlanguage font property instead.  If
+    $FONT already has #Mlanguage property, $LANGUAGE is ignored.
+
     @return
     This function returns a plist whose keys are family names and
     values are pointers to the object MFont.  The plist must be freed
@@ -2779,15 +2869,21 @@ mfont_resize_ratio (MFont *font)
 /***ja
     @brief եȤΥꥹȤ
 
-    ؿ mfont_list () ϥե졼 $FRAME ѲǽʥեȤΥꥹȤ֤
-    $FONT  NULL ǤʤС$FONT ȹפѲǽʥեȤΥꥹȤ֤
-    $LANGUAGE  @c Mnil ǤʤС$LANGUAGE 򥵥ݡȤѲǽʥեȤΥꥹȤ֤
-    $MAXNUM ϡ0 礭ˤϡ֤եȤοξ¤Ǥ롣
+    ؿ mfont_list () ϥե졼 $FRAME ѲǽʥեȤΥꥹȤ
+    ֤$FONT  NULL ǤʤС$FONT ȹפѲǽʥե
+    ΥꥹȤ֤$LANGUAGE  @c Mnil ǤʤС$LANGUAGE 򥵥ݡ
+    ȤѲǽʥեȤΥꥹȤ֤$MAXNUM ϡ0 礭
+    ˤϡ֤եȤοξ¤Ǥ롣
+
+     $LANGUAGE ϵǤȤΤˤꡢλѤ
+    ʤեȤ #Mlanguage ץѥƥȤ٤Ǥ롣
+     $FONT ǤˤΥץѥƥäƤ顢 $LANGUAGE ̵
+
 
     @return 
-    δؿϥեȥեߥ̾Ǥͤ MFont ֥ȤؤΥݥ󥿤Ǥ褦
-    plist ֤plist  m17n_object_unref () 
-    ǲɬפ롣եȤĤʤNULL ֤  */
+    δؿϥեȥեߥ̾Ǥͤ MFont ֥Ȥؤ
+    ݥ󥿤Ǥ褦plist ֤plist  m17n_object_unref () 
+    ɬפ롣եȤĤʤNULL ֤  */
 
 MPlist *
 mfont_list (MFrame *frame, MFont *font, MSymbol language, int maxnum)
@@ -2810,13 +2906,8 @@ mfont_list (MFrame *frame, MFont *font, MSymbol language, int maxnum)
     }
 
   if (language != Mnil)
-    {
-      /* ":lang=XXX" */
-      char *buf = alloca (MSYMBOL_NAMELEN (language) + 7);
-
-      sprintf (buf, ":lang=%s", MSYMBOL_NAME (language));
-      spec.capability = msymbol (buf);
-    }
+    spec.capability = merge_capability (spec.capability, Mlanguage, language,
+					0);
 
   font_list = mfont__list (frame, &spec, &spec, 0);
   if (! font_list)
@@ -2839,13 +2930,40 @@ mfont_list (MFrame *frame, MFont *font, MSymbol language, int maxnum)
   return plist;
 }
 
+/***en
+    @brief Get a list of font famiy names.
+
+    The mfont_list_family_names () functions returns a list of font
+    family names available on frame $FRAME.
+
+    @return
+
+    This function returns a plist whose keys are #Msymbol and values
+    are symbols representing font family names.  The elements are
+    sorted by alphabetical order.  The plist must be freed by
+    m17n_object_unref ().  If not font is found, it returns NULL.  */
+
+MPlist *
+mfont_list_family_names (MFrame *frame)
+{
+  MPlist *plist = mplist (), *p;
+
+  MPLIST_DO (p, frame->font_driver_list)
+    {
+      MFontDriver *driver = MPLIST_VAL (p);
+
+      (driver->list_family_names) (frame, plist);
+    }
+  return plist;
+}
+
 
 /*=*/
 
 /***en
     @brief Check the usability of a font.
 
-    The function mfont_check () checkes if $FONT can be used for
+    The mfont_check () function checkes if $FONT can be used for
     $SCRIPT and $LANGUAGE in $FONTSET on $FRAME.
 
     @return If the font is usable, return 1.  Otherwise return 0.
@@ -2867,6 +2985,117 @@ mfont_check (MFrame *frame, MFontset *fontset,
   return (score == 0 ? 2 : 1);
 }
 
+/*=*/
+
+/***en
+    @brief Check is a font matches with a font spec.
+
+    The mfont_match_p () function checks if $FONT matches with the
+    font-spec $SPEC.
+
+    @return 
+    If the font matches, 1 is returned.  Otherwise 0 is returned.  */
+
+int
+mfont_match_p (MFont *font, MFont *spec)
+{
+  return mfont__match_p (font, spec, MFONT_REGISTRY);
+}
+
+/*=*/
+/***en
+    @brief Open a font.
+
+    The mfont_open () function opens $FONT on $FRAME, and returns a
+    realized font.
+
+    @return
+    If the font was successfully opened, a realized font is returned.
+    Otherwize NULL is returned.
+
+    @seealso
+    mfont_close ().  */
+
+
+MFont *
+mfont_open (MFrame *frame, MFont *font)
+{
+  enum MFontType font_type = font->type;
+
+  if (font_type == MFONT_TYPE_SPEC)
+    return mfont_find (frame, font, NULL, 0);
+  if (font_type == MFONT_TYPE_OBJECT)
+    return (MFont *) mfont__open (frame, font, font);
+  if (font_type == MFONT_TYPE_REALIZED)
+    return font;
+  MERROR (MERROR_FONT, NULL);
+}
+
+/*=*/
+/***en
+    @brief Encapusulate a font.
+
+    The mfont_encapsulate () functions realizes a font by
+    encapusulating data $DATA or type $DATA_TYPE on #FRAME.  Currently
+    $DATA_TAPE is #Mfontconfig or #Mfreetype, and $DATA points to an
+    object of FcPattern or FT_Face respectively.
+
+    @return
+    If the operation was successful, a realized font is returned.
+    Otherwise NULL is return.
+
+    @seealso
+    mfont_close ().  */
+
+
+MFont *
+mfont_encapsulate (MFrame *frame, MSymbol data_type, void *data)
+{
+  MPlist *p;
+
+  MPLIST_DO (p, frame->font_driver_list)
+    {
+      MFontDriver *driver = MPLIST_VAL (p);
+      MRealizedFont *rfont;
+
+      if (driver->encapsulate
+	  && (rfont = driver->encapsulate (frame, data_type, data)))
+	return (MFont *) rfont;
+    }
+
+  return NULL;
+}
+
+/*=*/
+/***en
+    @brief Close a font.
+
+    The mfont_close () function close a realized font $FONT.  $FONT
+    must be opened previously by mfont_open () or mfont_encapsulate
+    ().
+
+    @return
+    If the operation was successful, 0 is returned.  Otherwise, -1 is
+    returned.
+
+    @seealso
+    mfont_open (), mfont_encapsulate ().  */
+
+int
+mfont_close (MFont *font)
+{
+  enum MFontType font_type = font->type;
+  MRealizedFont *rfont;
+
+  if (font_type != MFONT_TYPE_REALIZED)
+    MERROR (MERROR_FONT, -1);
+  rfont = (MRealizedFont *) font;
+  if (rfont->encapsulating
+      && rfont->driver->close)
+    rfont->driver->close (rfont);
+  return 0;
+}
+
 /*** @} */
 
 /*** @addtogroup m17nDebug */
