@@ -749,10 +749,8 @@ mfont__lookup_fontset (MRealizedFontset *realized, MGlyph *g, int *num,
       MFont request = realized->request;
 
       if (script != Mlatin)
-	/* These are not appropriate for non-Latin scripts.  */
-	request.property[MFONT_FOUNDRY]
-	  = request.property[MFONT_FAMILY]
-	  = request.property[MFONT_REGISTRY] = 0;
+	/* This is not appropriate for non-Latin scripts.  */
+	request.property[MFONT_REGISTRY] = 0;
 
       per_script = mplist_get (realized->per_script, script);
       if (! per_script)
@@ -775,7 +773,7 @@ mfont__lookup_fontset (MRealizedFontset *realized, MGlyph *g, int *num,
 				      g, num, size)))
 	return rfont;
 
-      if (per_lang)
+      if (per_lang && *num > 1)
 	*num = 1;
       if (language == Mt)
 	{
@@ -795,7 +793,7 @@ mfont__lookup_fontset (MRealizedFontset *realized, MGlyph *g, int *num,
 					  g, num, size)))
 	    return rfont;
 
-	  if (per_lang)
+	  if (per_lang && *num > 1)
 	    *num = 1;
 	  /* Then try the above (3) */
 	  MPLIST_DO (plist, per_script)
@@ -888,11 +886,12 @@ mfontset__get_font (MFrame *frame, MFontset *fontset,
   if (best)
     *best = 0;
 
+  if (language == Mnil)
+    language = Mt;
+
   if (script != Mnil)
     {
       per_script = get_per_script (fontset, script);
-      if (language == Mnil)
-	language = Mt;
       if ((per_lang = mplist_get (per_script, language))
 	  && (rfont = get_font_from_group (frame, per_lang, font)))
 	{
@@ -928,7 +927,7 @@ mfontset__get_font (MFrame *frame, MFontset *fontset,
 	}
     }
 
-  if (language != Mnil)
+  if (language != Mt)
     MPLIST_DO (per_script, fontset->per_script)
       {
 	if ((per_lang = mplist_get (MPLIST_PLIST (per_script), language))
