@@ -310,7 +310,7 @@ win_callback (MInputContext *ic, MSymbol command)
       if (ic->im->language != Mnil)
 	mtext_put_prop (ic->preedit, 0, mtext_nchars (ic->preedit), Mlanguage,
 			ic->im->language);
-      if (ic->candidate_list)
+      if (ic->candidate_list && ic->candidate_show)
 	mtext_push_prop (ic->preedit, ic->candidate_from, ic->candidate_to,
 			 Mface, mface_reverse_video);
       if (mtext_nchars (ic->produced) == 0)
@@ -547,7 +547,7 @@ minput__win_init ()
     plist = mplist_add (plist, Minput_reset, (void *) win_callback);
   }
 #if 0
-  /* This will make a caller of minput_method_open() pazzled.  */
+  /* This will make the caller of minput_method_open() pazzled.  */
   minput_driver = &minput_gui_driver;
 #endif
 
@@ -688,7 +688,7 @@ MSymbol Mxim;
     and/or "S-" (Shift) in this order.
 
     For instance, if the keysym name is "a" and the event has Shift,
-    Meta, and Hyper modifiers, the resulting name is "H-M-A".
+    Meta, and Hyper modifiers, the resulting name is "M-H-A".
 
     At last, a symbol who has the name is returned.  */
 
@@ -716,7 +716,7 @@ MSymbol Mxim;
     (Control), "S-" (Shift) ν֤դ
     
     ȤСkeysym ̾ "a" ǥ٥Ȥ Shift, Meta, and Hyper 
-    ǥեƤС̾ "H-M-A" Ǥ롣
+    ǥեƤС̾ "M-H-A" Ǥ롣
 
     Ǹˤ̾ĥܥ֤*/
 
