@@ -276,7 +276,6 @@ mlocale__init ()
 {
   M_locale = msymbol_as_managing_key ("  locale");
 
-  Mlanguage = msymbol ("language");
   Mterritory = msymbol ("territory");
   Mcodeset = msymbol ("codeset");
 
@@ -310,11 +309,6 @@ mlocale__fini ()
 /*** @addtogroup m17nLocale */
 /*** @{ */
 
-/*=*/
-/***en The symbol whose name is "language".  */
-/***ja "language" Ȥ̾ĥܥ.  */
-MSymbol Mlanguage;
-
 /*=*/
 /***en The symbol whose name is "territory".  */
 /***ja  "territory" Ȥ̾ĥܥ. */
