@@ -42,8 +42,8 @@ extern "C"
 
 #define M17NLIB_MAJOR_VERSION 1
 #define M17NLIB_MINOR_VERSION 3
-#define M17NLIB_PATCH_LEVEL 1
-#define M17NLIB_VERSION_NAME "1.3.1"
+#define M17NLIB_PATCH_LEVEL 3
+#define M17NLIB_VERSION_NAME "1.3.3"
 
 extern void m17n_init_core (void);
 #define M17N_INIT() m17n_init_core ()
@@ -440,6 +440,8 @@ extern MText *mtext_from_data (const void *data, int nitems,
 /*=*/
 /*** @} */
 
+extern MSymbol Mlanguage;
+
 /*
  *  (5-2) Functions to manipulate M-texts.  They correspond to string
  *   manipulating functions in libc.
@@ -555,7 +557,7 @@ extern MPlist *mplist_deserialize (MText *mt);
 /***en
     @brief Flag bits to control text property.
 
-    The mtext_property () funciton accepts logical OR of these flag
+    The mtext_property () function accepts logical OR of these flag
     bits as an argument.  They control the behaviour of the created
     text property as described in the documentation of each flag
     bit.  */
