@@ -1,5 +1,5 @@
 /* m17n-gd.c -- implementation of the GUI API on GD Library.
-   Copyright (C) 2004
+   Copyright (C) 2004, 2005, 2006
      National Institute of Advanced Industrial Science and Technology (AIST)
      Registration Number H15PRO112
 
@@ -67,9 +67,28 @@ static MSymbol M_rgb;
 static void
 read_rgb_txt ()
 {
-  FILE *fp = fopen ("/usr/lib/X11/rgb.txt", "r");
+  FILE *fp;
   int r, g, b;
 
+  /* At first, support HTML 4.0 color names. */
+  msymbol_put (msymbol ("black"), M_rgb, (void *) 0x000000);
+  msymbol_put (msymbol ("silver"), M_rgb, (void *) 0xC0C0C0);
+  msymbol_put (msymbol ("gray"), M_rgb, (void *) 0x808080);
+  msymbol_put (msymbol ("white"), M_rgb, (void *) 0xFFFFFF);
+  msymbol_put (msymbol ("maroon"), M_rgb, (void *) 0x800000);
+  msymbol_put (msymbol ("red"), M_rgb, (void *) 0xFF0000);
+  msymbol_put (msymbol ("purple"), M_rgb, (void *) 0x800080);
+  msymbol_put (msymbol ("fuchsia"), M_rgb, (void *) 0xFF00FF);
+  msymbol_put (msymbol ("green"), M_rgb, (void *) 0x008000);
+  msymbol_put (msymbol ("lime"), M_rgb, (void *) 0x00FF00);
+  msymbol_put (msymbol ("olive"), M_rgb, (void *) 0x808000);
+  msymbol_put (msymbol ("yellow"), M_rgb, (void *) 0xFFFF00);
+  msymbol_put (msymbol ("navy"), M_rgb, (void *) 0x000080);
+  msymbol_put (msymbol ("blue"), M_rgb, (void *) 0x0000FF);
+  msymbol_put (msymbol ("teal"), M_rgb, (void *) 0x008080);
+  msymbol_put (msymbol ("aqua"), M_rgb, (void *) 0x00FFFF);
+
+  fp = fopen ("/usr/lib/X11/rgb.txt", "r");
   if (! fp)
     fp = fopen ("/usr/X11R6/lib/X11/rgb.txt", "r");
   if (! fp)
@@ -327,10 +346,28 @@ gd_render (MDrawWindow win, int x, int y,
 	      if (bmp[j] > 0)
 		{
 		  int pixel1 = pixel;
+#if HAVE_GD > 1
 		  int alpha = gdAlphaTransparent * (255 - bmp[j]) / 255;
 
 		  if (alpha > 0)
 		    pixel1 = gdImageColorResolveAlpha (img, r, g, b, alpha);
+#else
+		  int f = bmp[j] >> 5;
+
+		  if (f < 7)
+		    {
+		      int r1, g1, b1, color1;
+
+		      pixel1 = gdImageGetPixel (img, xoff, yoff);
+		      r1 = gdImageRed (img, pixel1);
+		      g1 = gdImageGreen (img, pixel1);
+		      b1 = gdImageBlue (img, pixel1);
+		      color1 = ((((r * f + r1 * (7 - f)) / 7) << 16)
+				| (((g * f + g1 * (7 - f)) / 7) << 8)
+				| ((b * f + b1 * (7 - f)) / 7));
+		      pixel1 = RESOLVE_COLOR (img, color1);
+		    }
+#endif
 		  gdImageSetPixel (img, xoff, yoff, pixel1);
 		}
 	  }
@@ -821,6 +858,8 @@ device_open (MFrame *frame, MPlist *param)
   frame->realized_face_list = realized_face_list;
   frame->realized_fontset_list = realized_fontset_list;
   face = mface_copy (mface__default);
+  mface_put_prop (face, Mfoundry, Mnil);
+  mface_put_prop (face, Mfamily, Mnil);
   mplist_push (param, Mface, face);
   M17N_OBJECT_UNREF (face);
   return 0;
