@@ -168,6 +168,7 @@ extern MSymbol Mregistry;
 extern MSymbol Msize;
 extern MSymbol Mresolution;
 extern MSymbol Mmax_advance;
+extern MSymbol Motf;
 extern MSymbol Mfontfile;
 
 extern MSymbol Mfontconfig;
@@ -211,7 +212,6 @@ extern int mfont_set_encoding (MFont *font,
 
 extern MFont *mfont_find (MFrame *frame, MFont *spec,
 			  int *score, int limited_size);
-
 extern MSymbol *mfont_selection_priority ();
 
 extern int mfont_set_selection_priority (MSymbol *keys);
@@ -220,12 +220,21 @@ extern int mfont_resize_ratio (MFont *font);
 
 extern MPlist *mfont_list (MFrame *frame, MFont *font, MSymbol language,
 			   int maxnum);
+extern MPlist *mfont_list_family_names (MFrame *frame);
 
 typedef struct MFontset MFontset;
 
 extern int mfont_check (MFrame *frame, MFontset *fontset,
 			MSymbol script, MSymbol language, MFont *font);
 
+extern int mfont_match_p (MFont *font, MFont *spec);
+
+extern MFont *mfont_open (MFrame *frame, MFont *font);
+
+extern MFont *mfont_encapsulate (MFrame *frame, MSymbol data_type, void *data);
+
+extern int mfont_close (MFont *font);
+
 /* end of font module */
 /*=*/
 
