@@ -838,12 +838,23 @@ typedef struct MLocale MLocale;
 
 /*=*/
 
-extern MSymbol Mlanguage;
 extern MSymbol Miso639_1, Miso639_2;
 extern MSymbol Mterritory;
 extern MSymbol Mmodifier;
 extern MSymbol Mcodeset;
 
+extern MPlist *mlanguage_list (void);
+
+extern MSymbol mlanguage_code (MSymbol language, int len);
+
+extern MSymbol mlanguage_name (MSymbol language);
+
+extern MText *mlanguage_text (MSymbol language);
+
+extern MPlist *mscript_list (void);
+
+extern MPlist *mscript_language_list (MSymbol script);
+
 extern MLocale *mlocale_set (int category, const char *locale);
 
 extern MSymbol mlocale_get_prop (MLocale *locale, MSymbol key);
@@ -1114,6 +1125,11 @@ extern MSymbol Minput_focus_move;
 extern MSymbol Minput_focus_in;
 extern MSymbol Minput_focus_out;
 
+/** Symbols describing input method command/variable.  */
+extern MSymbol Minherited;
+extern MSymbol Mcustomized;
+extern MSymbol Mconfigured;
+
 /***en
     @brief Structure of input method.
 
@@ -1359,11 +1375,8 @@ extern int minput_filter (MInputContext *ic, MSymbol key, void *arg);
 
 extern int minput_lookup (MInputContext *ic, MSymbol key, void *arg,
 			  MText *mt);
-
-extern void minput_set_spot (MInputContext *ic, int x, int y,
-			     int ascent, int descent, int fontsize,
-			     MText *mt, int pos);
-
+extern void minput_set_spot (MInputContext *ic, int x, int y, int ascent,
+			     int descent, int fontsize, MText *mt, int pos);
 extern void minput_toggle (MInputContext *ic);
 
 extern void minput_reset_ic (MInputContext *ic);
@@ -1372,17 +1385,27 @@ extern MText *minput_get_description (MSymbol language, MSymbol name);
 
 extern MPlist *minput_get_title_icon (MSymbol language, MSymbol name);
 
+extern MPlist *minput_get_command (MSymbol language, MSymbol name,
+				   MSymbol command);
+extern int minput_config_command (MSymbol language, MSymbol name,
+				  MSymbol command, MPlist *keyseq);
+extern MPlist *minput_get_variable (MSymbol language, MSymbol name,
+				    MSymbol variable);
+extern int minput_config_variable (MSymbol language, MSymbol name,
+				   MSymbol variable, MPlist *value);
+extern char *minput_config_file (void);
+
+extern int minput_save_config (void);
+
+/* obsolete functions */
 extern MPlist *minput_get_commands (MSymbol language, MSymbol name);
 
 extern int minput_assign_command_keys (MSymbol language, MSymbol name,
 				       MSymbol command, MPlist *keys);
-
 extern MPlist *minput_get_variables (MSymbol language, MSymbol name);
 
 extern int minput_set_variable (MSymbol language, MSymbol name,
 				MSymbol variable, void *value);
-
-
 /*=*/
 /***     @} */
 
