@@ -100,10 +100,12 @@ extern MPlist *mplist__from_alist (MPlist *plist);
 
 extern MPlist *mplist__from_string (unsigned char *str, int n);
 
-extern int mplist__serialize (MText *mt, MPlist *plist);
+extern int mplist__serialize (MText *mt, MPlist *plist, int pretty);
 
 extern MPlist *mplist__conc (MPlist *plist, MPlist *tail);
 
 extern void mplist__pop_unref (MPlist *plist);
 
+extern MPlist *mplist__assq (MPlist *plist, MSymbol key);
+
 #endif  /* _M17N_PLIST_H_ */
